use std::io::BufRead;
use std::sync::Arc;

use lightmotif::abc::Alphabet;
use lightmotif::abc::Dna;
use lightmotif::abc::Protein;
use lightmotif_io::error::Error;

use pyo3::exceptions::PyOSError;
use pyo3::exceptions::PyValueError;
use pyo3::prelude::*;
use pyo3::types::PyBytes;

use super::pyfile::PyFileRead;
use super::CountMatrixData;
use super::Motif;
use super::ScoringMatrixData;
use super::WeightMatrixData;

// --- Error handling ----------------------------------------------------------

fn convert_error(error: Error) -> PyErr {
    match error {
        Error::InvalidData => PyValueError::new_err("invalid data"),
        Error::Io(err) => Arc::into_inner(err)
            .map(PyErr::from)
            .unwrap_or_else(|| PyOSError::new_err("unknown error")),
        Error::Nom(err) => PyValueError::new_err(format!("failed to parse input: {}", err)),
    }
}

// --- JASPAR motif ------------------------------------------------------------

#[pyclass(module = "lightmotif.lib", extends = Motif)]
pub struct JasparMotif {
    #[pyo3(get)]
    description: Option<String>,
}

impl JasparMotif {
    fn convert(record: Result<lightmotif_io::jaspar::Record, Error>) -> PyResult<PyObject> {
        let record = record.map_err(convert_error)?;
        let name = record.id().to_string();
        let description = record.description().map(String::from);
        let counts = record.into();
        Python::with_gil(|py| {
            let mut motif = Motif::from_counts(py, counts)?;
            motif.name = Some(name);
            let init = PyClassInitializer::from(motif).add_subclass(JasparMotif { description });
            Ok(Py::new(py, init)?.to_object(py))
        })
    }

    fn convert16<A>(record: Result<lightmotif_io::jaspar16::Record<A>, Error>) -> PyResult<PyObject>
    where
        A: Alphabet,
        CountMatrixData: From<lightmotif::pwm::CountMatrix<A>>,
        WeightMatrixData: From<lightmotif::pwm::WeightMatrix<A>>,
        ScoringMatrixData: From<lightmotif::pwm::ScoringMatrix<A>>,
    {
        let record = record.map_err(convert_error)?;
        let name = record.id().to_string();
        let description = record.description().map(String::from);
        let counts = record.into_matrix();
        Python::with_gil(|py| {
            let mut motif = Motif::from_counts(py, counts)?;
            motif.name = Some(name);
            let init = PyClassInitializer::from(motif).add_subclass(JasparMotif { description });
            Ok(Py::new(py, init)?.to_object(py))
        })
    }
}

// --- UniPROBE motif ----------------------------------------------------------

#[pyclass(module = "lightmotif.lib", extends = Motif)]
pub struct UniprobeMotif {}

impl UniprobeMotif {
    fn convert<A>(record: Result<lightmotif_io::uniprobe::Record<A>, Error>) -> PyResult<PyObject>
    where
        A: Alphabet,
        WeightMatrixData: From<lightmotif::pwm::WeightMatrix<A>>,
        ScoringMatrixData: From<lightmotif::pwm::ScoringMatrix<A>>,
    {
        let record = record.map_err(convert_error)?;
        let name = record.id().to_string();
        let freqs = record.into_matrix();
        let weights = freqs.to_weight(None);
        Python::with_gil(|py| {
            let mut motif = Motif::from_weights(py, weights)?;
            motif.name = Some(name);
            let init = PyClassInitializer::from(motif).add_subclass(UniprobeMotif {});
            Ok(Py::new(py, init)?.to_object(py))
        })
    }
}

// --- TRANSFAC records --------------------------------------------------------

#[pyclass(module = "lightmotif.lib", extends = Motif)]
pub struct TransfacMotif {
    #[pyo3(get)]
    id: Option<String>,
    #[pyo3(get)]
    accession: Option<String>,
    #[pyo3(get)]
    description: Option<String>,
}

impl TransfacMotif {
    fn convert<A>(record: Result<lightmotif_io::transfac::Record<A>, Error>) -> PyResult<PyObject>
    where
        A: Alphabet,
        CountMatrixData: From<lightmotif::pwm::CountMatrix<A>>,
        WeightMatrixData: From<lightmotif::pwm::WeightMatrix<A>>,
        ScoringMatrixData: From<lightmotif::pwm::ScoringMatrix<A>>,
    {
        let record = record.map_err(convert_error)?;

        let name = record.name().map(String::from);
        let description = record.description().map(String::from);
        let id = record.id().map(String::from);
        let accession = record.accession().map(String::from);
        let counts = record
            .to_counts()
            .ok_or_else(|| PyValueError::new_err("invalid count matrix"))?;
        Python::with_gil(|py| {
            let mut motif = Motif::from_counts(py, counts)?;
            motif.name = name;
            let init = PyClassInitializer::from(motif).add_subclass(TransfacMotif {
                description,
                accession,
                id,
            });
            Ok(Py::new(py, init)?.to_object(py))
        })
    }
}

// --- Loader ------------------------------------------------------------------

#[pyclass(module = "lightmotif.lib")]
pub struct Loader {
    reader: Box<dyn Iterator<Item = PyResult<PyObject>> + Send>,
}

#[pymethods]
impl Loader {
    fn __iter__(slf: PyRef<Self>) -> PyRef<Self> {
        slf
    }

    fn __next__(mut slf: PyRefMut<Self>) -> Option<PyResult<PyObject>> {
        slf.reader.next()
    }
}

/// Load the motifs contained in a file.
#[pyfunction]
#[pyo3(signature = (file, format="jaspar", *, protein=false))]
pub fn load(file: Bound<PyAny>, format: &str, protein: bool) -> PyResult<Loader> {
    let py = file.py();
    let pathlike = py
        .import_bound(pyo3::intern!(py, "os"))?
        .call_method1("fsencode", (&file,));
    let b: Box<dyn BufRead + Send> = if let Ok(path) = pathlike {
        // NOTE(@althonos): In theory this is safe because `os.fsencode` encodes
        //                  the PathLike object into the OS prefered encoding,
        //                  which is was OsStr wants. In practice, there may be
        //                  some weird bugs if that encoding is incorrect, idk...
        let encoded = path.downcast::<PyBytes>()?;
        let s = unsafe { std::ffi::OsStr::from_encoded_bytes_unchecked(encoded.as_bytes()) };
        std::fs::File::open(s)
            .map(std::io::BufReader::new)
            .map(Box::new)?
    } else {
        PyFileRead::from_ref(&file)
            .map(std::io::BufReader::new)
            .map(Box::new)?
    };
    let reader: Box<dyn Iterator<Item = PyResult<PyObject>> + Send> = match format {
        "jaspar" if protein => {
            return Err(PyValueError::new_err(
                "cannot read protein motifs from JASPAR format",
            ))
        }
        "jaspar16" if protein => {
            Box::new(lightmotif_io::jaspar16::read::<_, Protein>(b).map(JasparMotif::convert16))
        }
        "transfac" if protein => {
            Box::new(lightmotif_io::transfac::read::<_, Protein>(b).map(TransfacMotif::convert))
        }
        "uniprobe" if protein => {
            Box::new(lightmotif_io::uniprobe::read::<_, Protein>(b).map(UniprobeMotif::convert))
        }
        "jaspar" => Box::new(lightmotif_io::jaspar::read(b).map(JasparMotif::convert)),
        "jaspar16" => {
            Box::new(lightmotif_io::jaspar16::read::<_, Dna>(b).map(JasparMotif::convert16))
        }
        "transfac" => {
            Box::new(lightmotif_io::transfac::read::<_, Dna>(b).map(TransfacMotif::convert))
        }
        "uniprobe" => {
            Box::new(lightmotif_io::uniprobe::read::<_, Dna>(b).map(UniprobeMotif::convert))
        }
        _ => return Err(PyValueError::new_err(format!("invalid format: {}", format))),
    };
    Ok(Loader { reader })
}
